/*
 * Copyright (c) 2015 Vasa Consulting LLC
 *
 * This file is part of TYPE61CHK.
 *
 * TYPE61CHK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TYPE61CHK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TYPE61CHK.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <dirent.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sysexits.h>
#include <unistd.h>
#include <gpgme.h>
#include "support.h"

static gpgme_error_t		err;
static gpgme_ctx_t			ctx;
static gpgme_engine_info_t	eng_info;
static gpgme_data_t			data_plaintext, data_cipher;
static char					homedir_template[] = "/tmp/type61chk-gpg.XXXXXX";
static char				   *homedir = NULL;

static _Bool import_key(char *openpgpkey);
static void rmtree(char *dirpath);

_Bool
init_gpgme()
{
	_Bool	retval = false;

	homedir = mkdtemp(homedir_template);

	if (homedir == NULL)
		return EX_SOFTWARE;

	gpgme_check_version(NULL);
	err = gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP);

	if (err != GPG_ERR_NO_ERROR)
		/* No OpenPGP support */
		return EX_SOFTWARE;

	err = gpgme_new(&ctx);

	if (err != GPG_ERR_NO_ERROR)
		return EX_SOFTWARE;

	err = gpgme_get_engine_info(&eng_info);

	if (err != GPG_ERR_NO_ERROR)
		return EX_SOFTWARE;

	err = gpgme_ctx_set_engine_info(ctx, GPGME_PROTOCOL_OpenPGP,
										  eng_info->file_name,
										  homedir);

	if (err != GPG_ERR_NO_ERROR)
		return EX_SOFTWARE;

	gpgme_set_armor(ctx, 1);

	return retval;
}

void
free_gpgme()
{
	gpgme_data_release(data_plaintext);
	gpgme_data_release(data_cipher);
	gpgme_release(ctx);
	rmtree(homedir);
}

void
rmtree(char *dirpath)
{
	DIR			   *dir;
	struct dirent  *dp;
	char		   *cwd = NULL;

	cwd = getcwd(NULL, MAXPATHLEN);

	if (cwd == NULL)
		goto out;

	dir = opendir(dirpath);

	if (dir == NULL)
		goto out;

	if (chdir(dirpath))
		goto out;

	while ((dp = readdir(dir)) != NULL)
	{
		if ((! strncmp(dp->d_name, ".", strlen(dp->d_name)))
			|| (! strncmp(dp->d_name, "..", strlen(dp->d_name))))
			continue;

		if (dp->d_type == DT_DIR)
			rmtree(dp->d_name);

		unlink(dp->d_name);
	}

	rmdir(dirpath);

out:
	if (dir != NULL)
		closedir(dir);

	/* back to original dir */
	chdir(cwd);

	if (cwd != NULL)
		free(cwd);
}

/* attempt encryption of message with user's openpgpkey */
char *
encrypt_with_gpgme(char *sender, char *msg, char *openpgpkey)
{
	char				   *retval = NULL, *tmp = NULL;
	ssize_t					chars_read = 0;
	ssize_t					chars_pos = 0;
	gpgme_key_t				recp[2] = { NULL, NULL };
	gpgme_encrypt_flags_t	flags = GPGME_ENCRYPT_ALWAYS_TRUST;
	gpgme_encrypt_result_t	result;

	if (! import_key(openpgpkey))
	{
		fprintf(stdout, "couldn't import key\n");
		return NULL;
	}

	err = gpgme_op_keylist_start(ctx, sender, 0);

	if (err != GPG_ERR_NO_ERROR)
		goto out;

	err = gpgme_op_keylist_next(ctx, &recp[0]);

	if (err != GPG_ERR_NO_ERROR)
		goto out;

	err = gpgme_data_new_from_mem(&data_plaintext, msg, strlen(msg), 0);

	if (err != GPG_ERR_NO_ERROR)
		goto out;

	err = gpgme_data_new(&data_cipher);

	if (err != GPG_ERR_NO_ERROR)
		goto out;

	err = gpgme_op_encrypt(ctx, recp, flags, data_plaintext,
								 data_cipher);

	if (err != GPG_ERR_NO_ERROR)
		goto out;

	result = gpgme_op_encrypt_result(ctx);

	if (result->invalid_recipients)
	{
		fprintf(stdout, "invalid recipients found: '%s'\n",
				result->invalid_recipients->fpr);
		goto out;
	}

	gpgme_data_seek(data_cipher, 0, 0);
	retval = malloc(sizeof(char) * 64);

	if (retval == NULL)
		goto out;

	memset(retval, '\0', 64);

	while ((chars_read = gpgme_data_read(data_cipher,
										 retval + chars_pos,
										 64)))
	{
		chars_pos += chars_read;

		tmp = realloc(retval, (sizeof(char) * chars_pos) + 64);

		if (tmp != NULL)
			retval = tmp;

		memset(retval + chars_pos, '\0', sizeof(char) * 64);
	}

out:

	if (recp[0] != NULL)
		gpgme_key_unref(recp[0]);

	return retval;
}

static _Bool
import_key(char *openpgpkey)
{
	_Bool				retval = false;
	gpgme_data_t		keydata;
	gpgme_data_type_t	type;

	err = gpgme_data_new_from_mem(&keydata, openpgpkey, strlen(openpgpkey),
								  0);

	if (err != GPG_ERR_NO_ERROR)
		return false;

	type = gpgme_data_identify(keydata, 0);

	if (type != GPGME_DATA_TYPE_PGP_KEY)
		return false;

	err = gpgme_op_import(ctx, keydata);

	if (err != GPG_ERR_NO_ERROR)
		return false;
	else
		retval = true;

	gpgme_data_release(keydata);

	return retval;
}

