/*
 * Copyright (c) 2015 Vasa Consulting LLC
 *
 * This file is part of TYPE61CHK.
 *
 * TYPE61CHK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TYPE61CHK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TYPE61CHK.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <ldns/ldns.h>
#include "support.h"

static char *name_string = NULL;

static char *build_openpgpkey_rr_name(char *user_hashed, char *domain);

void
free_ldns(void)
{
	if (name_string)
		free(name_string);
}

/*
 * Return values:
 *
 * -3	Other problem
 * -2	No OPENPGPKEY record
 * -1	OPENPGPKEY record present and valid
 *  0	OPENPGPKEY record present and valid, DNSSEC-signed
 */
int
fetch_openpgpkey(char *user_hashed, char *domain, char **openpgpkey)
{
	int				retval = -1;
	ldns_resolver  *res = NULL;
	ldns_rdf	   *name = NULL;
	ldns_pkt	   *pkt = NULL;
	ldns_rr_list   *rr_list = NULL;
	ldns_rr		   *rr = NULL;
	ldns_status		status;

	name_string = build_openpgpkey_rr_name(user_hashed, domain);

	if (NULL == name_string)
	{
		retval = -3;
		goto out;
	}

	if (NULL == (name = ldns_dname_new_frm_str(name_string)))
	{
		retval = -3;
		goto out;
	}

	if (LDNS_STATUS_OK != (status = ldns_resolver_new_frm_file(&res, NULL)))
	{
		retval = -3;
		goto out;
	}

	ldns_resolver_set_dnssec(res, true);

	pkt = ldns_resolver_query(res,
							  name,
							  LDNS_RR_TYPE_OPENPGPKEY,
							  LDNS_RR_CLASS_IN,
							  LDNS_RD);

	if (NULL == pkt)
	{
		retval = -3;
		goto out;
	}

	rr_list = ldns_pkt_rr_list_by_type(pkt,
									   LDNS_RR_TYPE_OPENPGPKEY,
									   LDNS_SECTION_ANSWER);

	if (NULL == rr_list)
	{
		retval = -2;
		goto out;
	}
	else
	{
		rr = ldns_rr_list_rr(rr_list, 0);

		if (NULL == rr)
		{
			retval = -3;
			goto out;
		}

		/*
		 * NOTE: It's possible to have multiple OPENPGPKEY values under the
		 *       same name. But unlike with e.g. TLSA records, it doesn't
		 *       really make sense to do this, so ignore and use only the
		 *       first, albeit random, result.
		 */
		*openpgpkey = malloc(sizeof (char) * rr->_rdata_fields[0]->_size + 1);

		memcpy(*openpgpkey,
			   rr->_rdata_fields[0]->_data,
			   rr->_rdata_fields[0]->_size);

		openpgpkey[0][rr->_rdata_fields[0]->_size] = '\0';

		if (ldns_pkt_ad(pkt))
			retval = 0;
	}

out:
	if (rr_list)
		ldns_rr_list_deep_free(rr_list);

	if (pkt)
		ldns_pkt_free(pkt);

	if (res)
		ldns_resolver_deep_free(res);

	if (name)
		ldns_rdf_deep_free(name);

	return retval;
}

static char *
build_openpgpkey_rr_name(char *user_hashed, char *domain)
{
	if (NULL == user_hashed
		|| NULL == domain)
		return NULL;

	char   *middle = "._openpgpkey.";
	int		len = strlen(user_hashed) + strlen(middle) + strlen(domain) + 1;
	char   *retval = malloc(sizeof (char) * len);

	if (NULL != retval)
	{
		retval[len-1] = '\0';

		if (! snprintf(retval, len, "%s%s%s", user_hashed, middle, domain))
		{
			free(retval);
			return NULL;
		}

		return retval;
	}

	return NULL;
}

char *
query_text_ldns(void)
{
	char   *retval = NULL;
	retval = strndup(name_string, strlen(name_string));
	return retval;
}

