/*
 * Copyright (c) 2015 Vasa Consulting LLC
 *
 * This file is part of TYPE61CHK.
 *
 * TYPE61CHK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TYPE61CHK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TYPE61CHK.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Take sender address and perform DNSSEC OPENPGPKEY lookup
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sysexits.h>

#include "support.h"

#define T61_NAME			"TYPE61CHK"
#define T61_VERSION			"0.0.1"
#define T61_SHA224_BYTES	56

static int		usage(void);
static int		dichotomize(char *sender, char **user, char **domain);
static char	   *prep_msg(char *file, char *query);
static char	   *read_msg_file(char *file);
static char	   *replace_text(char *in, char *search, char *replace);

/* invocation */
char *proginv;
char *smtp_username = NULL, *smtp_password = NULL, *smtp_server = NULL;

int
main(int argc, char *argv[])
{
	int		retval = EX_OK;
	int		tmp;

	int		c = 0;
	char   *sender = NULL;

	char   *user = NULL;
	char   *domain = NULL;
	char   *user_hashed;
	char   *openpgpkey = NULL;
	_Bool	encrypt_reply = false;
	char   *msg_file = NULL;
	char   *msg = NULL;
	char   *final_msg = NULL;

	proginv = argv[0];

	while ((c = getopt(argc, argv, "e:hp:s:u:V")) != -1)
	{
		switch (c)
		{
			case 'e':
				smtp_server = optarg;
				break;

			case 'p':
				smtp_password = optarg;
				break;

			case 's':
				sender = optarg;
				break;

			case 'u':
				smtp_username = optarg;
				break;

			case 'V':
				printf("%s: %s v%s\n", argv[0], T61_NAME,
					   T61_VERSION);
				return EX_OK;
				break;

			case 'h':
			default:
				return usage();
				break;
		}
	}

	if (sender == NULL
		|| smtp_server == NULL
		|| smtp_username == NULL
		|| smtp_password == NULL
		|| optind != argc)
		return usage();

	init_curl(smtp_username);
	init_gpgme();
	init_sha();

	/* main */
	if ((retval = dichotomize(sender, &user, &domain)))
		return retval;

	user_hashed = malloc(sizeof(char) * T61_SHA224_BYTES + 1);
	memset(user_hashed, '\0', sizeof(char) * T61_SHA224_BYTES + 1);

	if ((retval = hash_user(user, user_hashed)))
		return retval;

	/*
	 * -3	Other problem
	 * -2	No OPENPGPKEY record
	 * -1	OPENPGPKEY record present and valid
	 *  0	OPENPGPKEY record present and valid, DNSSEC-signed
	 */
	tmp = fetch_openpgpkey(user_hashed, domain, &openpgpkey);

	if (tmp == 0)
	{
		encrypt_reply = true;
		msg_file = "messages/openpgpkey-record-signed.txt";
	}
	else if (tmp == -1)
	{
		encrypt_reply = true;
		msg_file = "messages/openpgpkey-record.txt";
	}
	else if (tmp == -2)
		msg_file = "messages/no-openpgpkey-record.txt";
	else
		retval = EX_SOFTWARE;

	if (retval != EX_SOFTWARE)
	{
		char *query = query_text_ldns();
		msg = prep_msg(msg_file, query);

		if (encrypt_reply)
		{
			final_msg = encrypt_with_gpgme(sender, msg, openpgpkey);

			if (final_msg == NULL)
				final_msg = msg;
		}
		else
			final_msg = msg;

		add_recp_to_curl(sender);

		if (add_msg_to_curl(final_msg, sender))
			return EX_SOFTWARE;

		send_email_via_curl();

		if (final_msg != msg)
			free(final_msg);

		free(msg);
		free(query);
	}

	if (openpgpkey != NULL)
		free(openpgpkey);

	if (user != NULL)
		free(user);

	if (domain != NULL)
		free(domain);

	if (user_hashed != NULL)
		free(user_hashed);

	free_curl();
	free_gpgme();

	return retval;
}

static int
usage()
{
	fprintf(stderr, "Usage: %s [options]\n"
					"\t-e server    \tSMTP server\n"
					"\t-p password  \tSMTP server password\n"
					"\t-s sender    \te-mail address to process\n"
					"\t-u username  \tSMTP server username\n"
					"\t-V           \tprint version and exit\n",
			 proginv);

	return EX_USAGE;
}

static int
dichotomize(char *sender, char **user, char **domain)
{
	char   *p;
	int		len = 0;

	p = strchr(sender, '@');
	if (p == NULL)
		return EX_DATAERR;

	len = strlen(sender);

	if ((*user = strndup(sender, (p - sender))) == NULL)
		return EX_OSERR;

	if ((*domain = strndup(p + 1, len - (p - sender) - 1)) == NULL)
		return EX_OSERR;

	return EX_OK;
}

static char *
prep_msg(char *file, char *query)
{
	char   *final	= NULL;
	char   *tmp		= NULL;

	final = read_msg_file(file);

	if (final == NULL)
		return NULL;

	tmp = replace_text(final, "{{query}}", query);

	if (tmp != NULL)
	{
		free(final);
		final = tmp;
	}

	return final;
}

static char *
read_msg_file(char *file)
{
	char   *msg = NULL;
	char   *tmp = NULL;
	FILE   *fp;
	size_t	total_nbr = 0, nbr; /* num bytes read */
	size_t	btr = 32; /* bytes to read */

	fp = fopen(file, "r");

	if (fp == NULL)
	{
		fprintf(stdout, "no fp, check for files in messages/ dir");
		return NULL;
	}

	msg = malloc(sizeof(char) * btr);
	memset(msg, '\0', sizeof(char) * btr);

	if (msg == NULL)
		return NULL;

	while (! feof(fp))
	{
		nbr = fread(msg + total_nbr, sizeof(char), btr, fp);
		total_nbr += nbr;

		/* If we asked for X bytes and got X bytes, need to realloc */
		if (nbr == btr)
		{
			tmp = realloc(msg, total_nbr + (sizeof(char) * btr));
			if (tmp == NULL)
			{
				free(msg);
				return NULL;
			}

			if (tmp != msg)
				msg = tmp;

			memset(msg + total_nbr, '\0', sizeof(char) * btr);

			/* handle reaching EOF */
			if (nbr < btr)
			{
				btr = (btr - (btr - nbr));
				memset(msg + total_nbr - btr, '\0', btr);
			}
		}

		if (nbr < btr)
			break;
	}

	fclose(fp);

	return msg;
}

static char *
replace_text(char *in, char *search, char *replace)
{
	char		   *out = NULL;
	size_t			c, d, e;
	size_t			in_len, out_len;
	const size_t	max_len = 1024;
	char		   *buf = NULL;
	size_t			buf_len = 0;
	size_t			buf_pos = 0;

	if (in == NULL || search == NULL || replace == NULL)
		return NULL;

	in_len = strnlen(in, max_len);
	buf_len = strnlen(search, max_len);
	out_len = (sizeof(char) * strnlen(in, max_len))
				 - strnlen(search, max_len)
				 + strnlen(replace, max_len)
				 ;

	buf = malloc(sizeof(char) * buf_len);
	strncpy(buf, search, buf_len);

	/* find the starting location of search */
	for (c = 0; c < in_len; c++)
	{
		if (in[c] == buf[buf_pos])
		{
			buf_pos++;
			if (buf_pos == buf_len - 1)
				break;
		}
		else
			buf_pos = 0;
	}

	if (buf_pos == buf_len - 1)
		c -= buf_pos - 1;
	else
		/* no search match found */
		goto out;

	out = malloc(out_len);

	if (out == NULL)
		goto out;

	/* XXX: Test when strlen(replace) < strlen(search) */
	strncpy(out, in, in_len);
	memset(out + in_len, 'X', out_len - in_len);
	memset(out + in_len, '\0', out_len - in_len);
	/*
	out[out_len-1] = '\0';
	*/

	/* cp from in[search+strlen(search)] to in[strlen(in)] into new buffer */
	for (d = out_len - 1, e = in_len - 1;
		 d > (c + (strnlen(replace, max_len) - 1));
		 d--, e--)
		memset(&out[d], in[e], 1);

	/* and finally insert replacement text */
	for (d = c, e = 0; d < (c + strnlen(replace, max_len)); d++)
	{
		if (&replace[e] == '\0')
			break;

		memset(&out[d], replace[e], 1);
		e++;
	}

	out[out_len-1] = '\0';

out:
	if (buf != NULL)
		free(buf);

	return out;
}

